/* @(#) $Id: acld.h 1277 2017-10-21 20:48:35Z leres $ (LBL) */
#ifndef acld_h
#define acld_h

#ifdef __FreeBSD__
#include <kvm.h>
#endif

#include "cf.h"
#include "io.h"
#include "stats.h"
#include "timer.h"
#include "version.h"

/* Maximum number of "sent" replicant requests */
#define REPLICANT_DEPTH 100

/* acld states */
enum acldstate {
	ASTATE_NOTCONNECTED = 0,
	ASTATE_CONNECTED,
	ASTATE_LOGGEDIN,
	ASTATE_SENTLOGIN,
	ASTATE_SENTLISTACL,
	ASTATE_SENTLISTROUTE,
	ASTATE_READRESPONSE,
	ASTATE_READERROR,
	ASTATE_READACL,
	ASTATE_READROUTE,
	ASTATE_SENTATTR,
};

#define ACLDREADY(p) \
    ((p)->state == ASTATE_LOGGEDIN && \
    (p)->sentattrs && (p)->listedroutes && (p)->listedallacls)

/* Both the req and the optional replicant request are done */
#define REQUESTDONE(rp) \
    ((rp) != NULL && \
    (rp)->state == RSTATE_DONE && \
    ((rp)->rereq == NULL || (rp)->rereq->state == RESTATE_DONE))

/* Seconds to wait for acld to login to the router before giving up */
#define CHILD_REQUEST_TIMEOUT_SECS	30

/* State */
struct state {
	enum acldstate state;		/* ASTATE_* */
	time_t loginout_ts;		/* time we last logged in or out */
	const char *cmd;		/* command sent */
	int s;				/* IPv4 listen socket */
	int s6;				/* IPv6 listen socket */
	int ro;				/* IPv4 listen socket (read only) */
	int ro6;			/* IPv6 listen socket (read only) */
	int web;			/* IPv4 listen socket (web reg) */
	int web6;			/* IPv6 listen socket (web reg) */
	int rfd;			/* expect child read pipe */
	int wfd;			/* expect child read pipe */
	int refd;			/* replicant socket */
	u_int32_t recookie;		/* replicant cookie */
	int re_secs;			/* replicant connect retry seconds */
	int replicantdepth;		/* number of outstanding requests */
	struct iobuf re_rbuf;		/* replicant read buffer */
	struct iobuf re_wbuf;		/* replicant write buffer */
	struct rereq *new_rereq;	/* new replicant requests */
	struct rereq *sent_rereq;	/* pending replicant requests */

#ifdef __FreeBSD__
	int routefd;			/* FreeBSD routing raw socket */
	kvm_t *kvmd;			/* FreeBSD kvm descriptor */
#endif
	pid_t pid;			/* child pid */
	u_int nclients;			/* next client number */

	/*
	 * Submit all ready requests to the child
	 * Commit when there are no ready requests left
	 */
	int multiball;

	int listedallacls;		/* have we listed all the acl's? */
	int listedroutes;		/* have we listed the routes? */
	int sentattrs;			/* have we sent the attributes? */

	struct aroute *routes;		/* list of routes from router */
	struct array routes_array;

	int nullzerolen;		/* current number of nullzero routes */
	struct stats nullzerostats;	/* nullzero statistics */

	struct timeval timeout;		/* select timeout */
	struct iobuf rbuf;		/* child read buffer */
	struct req *req;		/* linked list of server requests */
	struct client *clients;		/* linked list of clients */
	struct cf *cf;			/* configuration */
	struct timer t_ayt;		/* timer for "ayt" */
	struct timer t_compact;		/* timer for "compact" */
	struct timer t_login;		/* timer for next login attempt */
	struct timer t_sync;		/* timer for next sync */
	struct timer t_re_connect;	/* timer for replicant connect */
	struct timer t_re_timeout;	/* timer for replicant session */
	struct timer t_child;		/* timer for generic child use */

	/* Child hooks */
	const char *(*f_childcheckreq)(struct req *);
	void (*f_childcommit)(struct state *);
	void (*f_childconnect)(struct state *);
	int (*f_childinput)(struct state *);
	void (*f_childkill)(struct state *);
	void (*f_childlistacl)(struct state *);
	void (*f_childlistroute)(struct state *);
	void (*f_childlogin)(struct state *);
	int (*f_childread)(int, struct iobuf *);
	int (*f_childready)(struct state *);
	void (*f_childrequest)(struct state *, struct aclgroup *, struct req *);
	void (*f_childsend)(struct state *, const char *, ...)
	    __attribute__ ((format (printf, 2, 3)));
	void (*f_childsendattr)(struct state *);

	/* Client hooks */
	void (*f_clientcompact)(struct state *, struct client *cl,
	    struct req *rp);
	void (*f_clientdroprestore)(struct state *, struct client *,
	    struct req *);
	void (*f_clientlistroute)(struct state *, struct client *,
	    struct req *, int);
	const char *(*f_clientmapaclname)(const char *, struct acl *);
	void (*f_clientnullzero)(struct state *, struct client *, struct req *);
	void (*f_clientquerynullzero)(struct state *, struct client *,
	    struct req *);

	/* Server hooks */
	void (*f_serverayt)(struct state *);
	void (*f_servercompact)(struct state *);
	void (*f_serversync)(struct state *);
};

#include "acl.h"
#include "client.h"
#include "route.h"

#include "whitelist.h"

/* Request types */
enum reqtype {
	REQ_UNKNOWN = 0,
	REQ_RELOAD,		/* kill expect child process & reload config */
	REQ_DROP,		/* block a host or subnet */
	REQ_RESTORE,		/* unblock a host or subnet */
	REQ_BLOCKHOSTHOST,	/* block a host to host pair */
	REQ_RESTOREHOSTHOST,	/* unblock a host to host pair */
	REQ_DROPUDPPORT,	/* block a udp port */
	REQ_RESTOREUDPPORT,	/* unblock a udp port */
	REQ_DROPTCPPORT,	/* block a tcp port */
	REQ_RESTORETCPPORT,	/* unblock a tcp port */
	REQ_DROPTCPSYNPORT,	/* block a tcp syn port */
	REQ_RESTORETCPSYNPORT,	/* unblock a tcp syn port */
	REQ_PERMITUDPDSTHOSTPORT, /* permit udp dst-host/port */
	REQ_UNPERMITUDPDSTHOSTPORT, /* remove permit udp dst-host/port */
	REQ_PERMITTCPDSTHOSTPORT, /* permit tcp dst-host/port */
	REQ_UNPERMITTCPDSTHOSTPORT, /* remove permit tcp dst-host/port */
	REQ_DROPTCPDSTHOSTPORT,	/* block tcp dst-host/port */
	REQ_RESTORETCPDSTHOSTPORT, /* remove tcp dst-host/port */
	REQ_NULLZERO,		/* create null zero host route */
	REQ_NONULLZERO,		/* remove null zero host route */
	REQ_STATE,		/* display current acld state */
	REQ_LISTACL,		/* list contents of specified acl */
	REQ_LISTROUTE,		/* list routes (with route type character) */
	REQ_WHITELIST,		/* list whitelist */
	REQ_ADDWHITELIST,	/* add whitelist address */
	REQ_REMWHITELIST,	/* remove whitelist address */
	REQ_QUERY,		/* check host, host to host or subnet block */
	REQ_QUERYNULLZERO,	/* check for a nullzero host route */
	REQ_QUERYWHITELIST,	/* check for a whitelist host/net */
	REQ_LISTNULLZERO,	/* list nullzero routes */
	REQ_COMPACT,		/* schedule compaction of specified acl */
	REQ_TEST,		/* (undocumented) */
	REQ_EXIT,		/* terminate client session */
	REQ_HELP,		/* display brief help message */
	REQ_SYNC,		/* copy router config to nonvolatile memory */
	REQ_REPLICANT,		/* register as a replicant provider */
	REQ_AYT,		/* check that router is still alive */
};

/* Include after route.h and enum reqtype */
#include "replicant.h"

/* Request states */
enum reqstate {
	RSTATE_PENDING = 0,	/* not processed yet */
	RSTATE_READCOMMENT,	/* reading client comments */
	RSTATE_READRESPONSE,	/* reading child response */
	RSTATE_CHILD,		/* waiting for child to reply */
	RSTATE_DONE		/* ok to remove from queue */
};

/* Request flags */
enum reqflags {
	RFLAG_CONTINUE	= 0x1,	/* request has a payload */
	RFLAG_FAILED	= 0x2,	/* request resulted in a failed status */
	RFLAG_IGNORE	= 0x4,	/* request didn't result in any changes */
};

/* Request */
struct req {
	struct req *next;
	struct rereq *rereq;		/* link to replicant request */
	enum reqtype type;		/* REQ_* */
	enum reqstate state;		/* RSTATE_* */
	enum reqflags flags;		/* request flags */
	const char *cmd;		/* client command name */
	struct aclgroup *aclgroup;	/* aclgroup pointer */
	char *aclname;			/* listacl, compact, drop udp/tcp ACL */
	u_int32_t cookie;
	u_int32_t multiball;		/* multiball request */
	struct timeval ats;		/* arrival timestamp */
	struct timeval cts;		/* completion timestamp */
	int an;				/* argument count */
	char **av;			/* argument vector */
	char *comment;
	struct iobuf payload;
	struct aroute nullzero;
	struct addr whitelist;
	struct acl acl;
};

#define LOGIN_SECS		15
#define AYT_SECS		(10 * 60)
#define SYNC_SECS		(30 * 60)
#define SELECT_SECS		15
#define REPLICANT_SESSION_SECS	45	/* half of acl.exp's timeout */

/* Must be greater than SELECT_SECS */
#define IDLE_SECS		(2 * SELECT_SECS)

#define MAX_IPV4_WIDTH		16	/* widest IPv4 cidr block */
#define MAX_IPV6_WIDTH		32	/* widest IPv6 cidr block */

#define DEFAULT_IPV4_WIDTH	24	/* default IPv4 cidr block width */
#define DEFAULT_IPV6_WIDTH	64	/* default IPv6 cidr block width */

/* https://docs.freebsd.org/info/cpp/cpp.info.Stringification.html */
#define _HERESTR1(fn, ln) _HERESTR2(fn, ln)
#define _HERESTR2(fn, ln) fn ":" #ln
#define HERESTR _HERESTR1(__FILE__, __LINE__)

/* Globals */
extern int debug;
extern int verbose;
extern int reload;
extern const char *prog;
extern struct s2v astate2str[];
extern struct s2v rstate2str[];
extern const char *configfn;
extern const char *logfile;
extern char hostname[];

/* Functions */
extern void appendreq(struct req **, struct req *);
extern void freereq(struct state *, struct req *);
extern int getastatesecs(struct state *);
extern struct client *getreqclient(struct state *, struct req *);
extern struct req *newreq(const char *);
extern void setastate(struct state *, enum acldstate);
#endif
