/*
 * Copyright (c) 2016, 2017
 * The Regents of the University of California. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: jreq.c 1273 2017-08-08 19:06:50Z leres $ (LBL)";
#endif

/* Generate Junoscript configuration RPCs */

#include <sys/types.h>

#include <libxml/encoding.h>
#include <libxml/xmlwriter.h>

#include <stdio.h>
#include <string.h>

#include "acld.h"
#include "jreq.h"

/* Globals */
static int depth;				/* current XML depth */
static int cdepth;				/* "configuration" XML depth */
static xmlTextWriterPtr xp;
static xmlBufferPtr jreq_buf;
static char currentprefix[64];			/* currently configuring this */

/* Forwards */
static const char *addnamedelement(const char *, const char *);
static const char *addnamedelementwithattr(const char *, const char *,
    const char *, const char *);
static const char *climbup(int);
static const char *jreq_add_firewall(struct jrequest *);
static const char *jreq_add_prefix(struct jrequest *);
static const char *jreq_add_prefix_addr(struct jrequest *);
static const char *jreq_start(void);
static const char *startchild(const char *);
static const char *startchildwithattr(const char *, const char *, const char *);

static const char *
addnamedelement(const char *tag, const char *name)
{
	int rc;

	rc = xmlTextWriterWriteElement(xp, (xmlChar *)tag, (xmlChar *)name);
	if (rc < 0)
		return ("addnamedelement: xmlTextWriterWriteElement failed");
	return (NULL);
}

static const char *
addnamedelementwithattr(const char *tag, const char *name,
    const char *key, const char *val)
{
	int rc;
	const char *errmsg;

	rc = xmlTextWriterStartElement(xp, (xmlChar *)tag);
	if (rc < 0) {
		errmsg =
		    "addnamedelementwithattr: xmlTextWriterStartElement failed";
		return (errmsg);
	}
	++depth;
	rc = xmlTextWriterWriteAttribute(xp, (xmlChar *)key, (xmlChar *)val);
	if (rc < 0) {
		errmsg = "addnamedelementwithattr:"
		    " xmlTextWriterWriteAttribute failed";
		return (errmsg);
	}
	rc = xmlTextWriterWriteString(xp, (xmlChar *)name);
	if (rc < 0) {
		errmsg =
		    "addnamedelementwithattr: xmlTextWriterWriteString failed";
		return (errmsg);
	}

	return (NULL);
}

/* Climb back to depth2 */
static const char *
climbup(int depth2)
{
	int rc;

	if (depth2 < 0)
		depth2 = 0;
	while (depth > depth2) {
		rc = xmlTextWriterEndElement(xp);
		if (rc < 0)
			return ("endchild: xmlTextWriterEndElement failed");
		--depth;
	}
	return (NULL);
}

/* Add a firewall change */
static const char *
jreq_add_firewall(struct jrequest *jp)
{
	int dodelete, depth3;
	const char *thenaction, *prefixdirection, *errmsg;

	/* XXX need to delete port firewall config if removing last port */

	thenaction = "discard";
	prefixdirection = "prefix-list";
	dodelete = 0;
	switch (jp->type) {
	case REQ_RESTORETCPPORT:
	case REQ_RESTOREUDPPORT:
	case REQ_RESTORETCPSYNPORT:
		++dodelete;
		break;

	case REQ_PERMITUDPDSTHOSTPORT:
	case REQ_PERMITTCPDSTHOSTPORT:
		thenaction = "accept";
		prefixdirection = "destination-prefix-list";
		break;

	default:
		break;
	}

	switch (jp->type) {

	case REQ_DROP:
	case REQ_DROPUDPPORT:
	case REQ_DROPTCPPORT:
	case REQ_RESTORETCPPORT:
	case REQ_RESTOREUDPPORT:
	case REQ_DROPTCPSYNPORT:
	case REQ_RESTORETCPSYNPORT:
	case REQ_PERMITUDPDSTHOSTPORT:
	case REQ_PERMITTCPDSTHOSTPORT:
		/* We need to start at configuration depth */
		if (depth != cdepth) {
			errmsg = climbup(cdepth);
			if (errmsg != NULL)
				return (errmsg);
		}
		currentprefix[0] = '\0';

		errmsg = startchild("firewall");
		if (errmsg != NULL)
			return (errmsg);
		errmsg = startchild("family");
		if (errmsg != NULL)
			return (errmsg);
		errmsg = startchild(jp->family);
		if (errmsg != NULL)
			return (errmsg);
		errmsg = startchild("filter");
		if (errmsg != NULL)
			return (errmsg);
		errmsg = addnamedelement("name", jp->filter);
		if (errmsg != NULL)
			return (errmsg);

		if (!dodelete) {
			errmsg = startchild("interface-specific");
			if (errmsg != NULL)
				return (errmsg);
			errmsg = climbup(depth - 1);
			if (errmsg != NULL)
				return (errmsg);
		}


		if (!jp->doremoveterm)
			errmsg = startchild("term");
		else
			errmsg = startchildwithattr("term", "delete", "delete");

		if (errmsg != NULL)
			return (errmsg);
		errmsg = addnamedelement("name", jp->seq);
		if (errmsg != NULL)
			return (errmsg);

		/* This is all we need to do when deleting the term */
		if (jp->doremoveterm)
			break;

		depth3 = depth;
		errmsg = startchild("from");
		if (errmsg != NULL)
			return (errmsg);

		/* Add address if present */
		if (jp->addr[0] != '\0') {
			errmsg = startchild(prefixdirection);
			if (errmsg != NULL)
				return (errmsg);
			errmsg = addnamedelement("name", jp->prefix);
			if (errmsg != NULL)
				return (errmsg);
			errmsg = climbup(depth - 1);
			if (errmsg != NULL)
				return (errmsg);
		}

		/* Add protocol if present (and not deleting a port) */
		if (!dodelete && jp->proto[0] != '\0') {
			errmsg = addnamedelement(jp->protokey, jp->proto);
			if (errmsg != NULL)
				return (errmsg);
		}

		/* Add port if present */
		if (jp->port[0] != '\0') {
			if (!dodelete)
				errmsg = addnamedelement("destination-port",
				    jp->port);
			else
				errmsg = addnamedelementwithattr(
				    "destination-port",
				    jp->port, "delete", "delete");
			if (errmsg != NULL)
				return (errmsg);
		}

		/* Add optional syn requirement */
		if (jp->type == REQ_DROPTCPSYNPORT) {
			errmsg = startchild("tcp-initial");
			if (errmsg != NULL)
				return (errmsg);
			errmsg = climbup(depth - 1);
			if (errmsg != NULL)
				return (errmsg);
		}

		errmsg = climbup(depth3);
		if (errmsg != NULL)
			return (errmsg);
		if (!dodelete) {
			errmsg = startchild("then");
			if (errmsg != NULL)
				return (errmsg);

			/* Want "<discard></discard>" instead of "<discard/>" */
			errmsg = addnamedelement(thenaction, "");
			if (errmsg != NULL)
				return (errmsg);
		}
		break;

	case REQ_RESTORE:
	case REQ_UNPERMITTCPDSTHOSTPORT:
	case REQ_UNPERMITUDPDSTHOSTPORT:
		/* We don't currently try and clean up unused filter terms */
		break;

	default:
		abort();
	}
	return (NULL);
}

static const char *
jreq_add_prefix(struct jrequest *jp)
{
	const char *errmsg;

	/* Optimize for updating prefix lists */
	if (!jp->dofirewall &&
	    jp->addr[0] != '\0' &&
	    jp->prefix[0] != '\0' &&
	    strcmp(currentprefix, jp->prefix) == 0) {
		errmsg = jreq_add_prefix_addr(jp);
		if (errmsg != NULL)
			return (errmsg);
		return (NULL);
	}

	switch (jp->type) {

	case REQ_DROP:
	case REQ_RESTORE:
	case REQ_PERMITUDPDSTHOSTPORT:
	case REQ_PERMITTCPDSTHOSTPORT:
	case REQ_UNPERMITTCPDSTHOSTPORT:
	case REQ_UNPERMITUDPDSTHOSTPORT:
		/* We need to start at configuration depth */
		if (depth != cdepth) {
			errmsg = climbup(cdepth);
			if (errmsg != NULL)
				return (errmsg);
		}
		currentprefix[0] = '\0';

		/* Prefix list change */
		errmsg = startchild("policy-options");
		if (errmsg != NULL)
			return (errmsg);
		errmsg = startchild("prefix-list");
		if (errmsg != NULL)
			return (errmsg);
		errmsg = addnamedelement("name", jp->prefix);
		if (errmsg != NULL)
			return (errmsg);
		strlcpy(currentprefix, jp->prefix, sizeof(currentprefix));

		errmsg = jreq_add_prefix_addr(jp);
		if (errmsg != NULL)
			return (errmsg);
		break;

	case REQ_DROPUDPPORT:
	case REQ_DROPTCPPORT:
	case REQ_RESTORETCPPORT:
	case REQ_RESTOREUDPPORT:
	case REQ_DROPTCPSYNPORT:
	case REQ_RESTORETCPSYNPORT:
		/* Ports don't use prefix-lists */
		break;

	default:
		abort();
	}
	return (NULL);
}

static const char *
jreq_add_prefix_addr(struct jrequest *jp)
{
	const char *errmsg;

	switch (jp->type) {

	case REQ_RESTORE:
	case REQ_UNPERMITTCPDSTHOSTPORT:
	case REQ_UNPERMITUDPDSTHOSTPORT:
		errmsg = startchildwithattr("prefix-list-item",
		    "delete", "delete");
		break;
	default:
		errmsg = startchild("prefix-list-item");
	}
	if (errmsg != NULL)
		return (errmsg);
	errmsg = addnamedelement("name", jp->addr);
	if (errmsg != NULL)
		return (errmsg);
	errmsg = climbup(depth - 1);
	if (errmsg != NULL)
		return (errmsg);
	return (NULL);
}

const char *
jreq_add(struct jrequest *jp)
{
	const char *errmsg;

	/* Start the configuration if we haven't already */
	if (xp == NULL) {
		errmsg = jreq_start();
		if (errmsg != NULL)
			return (errmsg);
	}

	/* Only add firewall config if necessary */
	if (jp->dofirewall) {
		errmsg = jreq_add_firewall(jp);
		if (errmsg != NULL)
			return (errmsg);
	}

	errmsg = jreq_add_prefix(jp);
	if (errmsg != NULL)
		return (errmsg);

	return (NULL);
}

/* Returns true if there is XML that hasn't been rendered */
int
jreq_dirty()
{
	return (xp != NULL);
}

void
jreq_cleanup()
{
	if (xp != NULL) {
		xmlFreeTextWriter(xp);
		xp = NULL;
	}
	if (jreq_buf != NULL) {
		xmlBufferFree(jreq_buf);
		jreq_buf = NULL;
	}
	depth = 0;
	cdepth = 0;
	currentprefix[0] = '\0';
}

void
jreq_init()
{
	depth = 0;
	if (xp != NULL)
		jreq_cleanup();
}

const char *
jreq_render(const char **pp)
{
	int rc;

	*pp = NULL;
	rc = xmlTextWriterEndDocument(xp);
	if (rc < 0)
		return ("Error at xmlTextWriterEndDocument");

	*pp = (const char *)jreq_buf->content;
	return (NULL);
}

static const char *
jreq_start()
{
	int rc;
	const char *errmsg;

	/* Create the XML buffer */
	jreq_buf = xmlBufferCreate();
	if (jreq_buf == NULL)
		return ("Error creating the xml buffer");

	/* Create the XML writer */
	xp = xmlNewTextWriterMemory(jreq_buf, 0);
	if (xp == NULL)
		return ("Error creating the xml writer");

	/* Indent output */
	rc = xmlTextWriterSetIndent(xp, 1);
	if (rc < 0)
		return ("Error at xmlTextWriterSetIndent");

	/* Indent using 4 blanks */
	rc = xmlTextWriterSetIndentString(xp, (const xmlChar *)"    ");
	if (rc < 0)
		return ("Error at xmlTextWriterSetIndentString");

	depth = 0;

	errmsg = startchild("rpc");
	if (errmsg != NULL)
		return (errmsg);
	errmsg = startchildwithattr("load-configuration", "action", "merge");
	if (errmsg != NULL)
		return (errmsg);
	errmsg = startchild("configuration");
	if (errmsg != NULL)
		return (errmsg);

	/* Save "configuration" depth */
	cdepth = depth;
	return (NULL);
}

static const char *
startchild(const char *tag)
{
	int rc;

	rc = xmlTextWriterStartElement(xp, (xmlChar *)tag);
	if (rc < 0)
		return ("startchild: xmlTextWriterStartElement failed");
	++depth;
	return (NULL);
}

static const char *
startchildwithattr(const char *tag, const char *key, const char *val)
{
	int rc;
	const char *errmsg;

	rc = xmlTextWriterStartElement(xp, (xmlChar *)tag);
	if (rc < 0) {
		errmsg = "startchildwithattr: xmlTextWriterStartElement failed";
		return (errmsg);
	}
	rc = xmlTextWriterWriteAttribute(xp, (xmlChar *)key, (xmlChar *)val);
	if (rc < 0) {
		errmsg =
		    "startchildwithattr: xmlTextWriterWriteAttribute failed";
		return (errmsg);
	}
	++depth;
	return (NULL);
}
